"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.CustomCostAnalysis = void 0;
const graphql_1 = require("graphql");
const costAnalysis_1 = __importDefault(require("graphql-cost-analysis/dist/costAnalysis"));
const graphql_depth_limit_1 = __importDefault(require("graphql-depth-limit"));
const gql_loader_1 = __importDefault(require("../loaders/gql.loader"));
const logger_1 = __importDefault(require("../logger"));
class CustomCostAnalysis extends costAnalysis_1.default {
    constructor(context, options) {
        super(context, options);
        this.context = context;
        this.options = options;
        this.cost = 1;
    }
}
exports.CustomCostAnalysis = CustomCostAnalysis;
const costAnalysisTool = async (config, requestCostTool) => {
    var _a, _b, _c, _d, _e, _f;
    const maxDepth = ((_b = (_a = config.security) === null || _a === void 0 ? void 0 : _a.graphql) === null || _b === void 0 ? void 0 : _b.maxDepth) || 10;
    const maximumCost = ((_d = (_c = config.security) === null || _c === void 0 ? void 0 : _c.graphql) === null || _d === void 0 ? void 0 : _d.maxCost) || 1000;
    const userCostPerHour = ((_f = (_e = config.security) === null || _e === void 0 ? void 0 : _e.graphql) === null || _f === void 0 ? void 0 : _f.userCostPerHour) || 300000;
    const hourInMs = 1000 * 60 * 60;
    const costPerMs = userCostPerHour / hourInMs;
    const roundTo = (value, to) => Math.round(value * (10 ** to)) / (10 ** to);
    const schemaString = await gql_loader_1.default.getTypeDefs();
    const schema = graphql_1.buildSchema(schemaString);
    const sendError = (res, error, status = 400) => {
        res.status(status);
        res.json(error);
    };
    async function middleware(req, res, next) {
        const { body } = req;
        if (req.method !== 'POST' || req.body.operationName === 'IntrospectionQuery') {
            return next();
        }
        const query = graphql_1.parse(body.query);
        const typeInfo = new graphql_1.TypeInfo(schema);
        const validationContext = new graphql_1.ValidationContext(schema, query, typeInfo, () => { });
        let depth = 0;
        const setDepth = (depths) => { depth = Math.max(...Object.values(depths)); };
        graphql_depth_limit_1.default(maxDepth, {}, setDepth)(validationContext);
        if (Number.isNaN(depth)) {
            return sendError(res, new graphql_1.GraphQLError(`The query exceeds maximum operation depth of ${maxDepth}`));
        }
        if (!Number.isFinite(maximumCost)) {
            return next();
        }
        const visitor = new CustomCostAnalysis(validationContext, {
            maximumCost,
            defaultCost: 1,
            complexityRange: { min: 1, max: 100 },
        });
        graphql_1.visit(query, graphql_1.visitWithTypeInfo(typeInfo, visitor));
        const { cost } = visitor;
        if (cost > maximumCost) {
            return sendError(res, new graphql_1.GraphQLError(`The query exceeds maximum complexity of ${maximumCost}`));
        }
        requestCostTool.setCost(req, cost);
        const totalCost = await requestCostTool.getTotalCost(req) + cost;
        if (totalCost > userCostPerHour) {
            return sendError(res, new graphql_1.GraphQLError(`Too many requests. Exceeded complexity limit of ${userCostPerHour} per hour`), 429);
        }
        return next();
    }
    const plugin = {
        requestDidStart({ context }) {
            const request = context.request;
            let start;
            return {
                responseForOperation: async () => {
                    await requestCostTool.saveCost(request);
                    start = Date.now();
                    return null;
                },
                willSendResponse: async () => {
                    const end = Date.now();
                    if (!request.costData) {
                        return;
                    }
                    const { body, cost } = request.costData;
                    const serverTime = end - start;
                    const expectedCost = roundTo(costPerMs * serverTime, 2);
                    const expectedTime = roundTo(costPerMs * cost, 2);
                    if (cost < expectedCost) {
                        let warning = `Request with cost ${cost} was lasting for ${serverTime}ms but expected less than ${expectedTime}ms.\n`;
                        warning += `Expected cost for ${serverTime}ms is ${expectedCost}.\n`;
                        warning += `The request:\n${JSON.stringify(body)}`;
                        logger_1.default.warn(warning, 'GraphQL Security');
                    }
                },
            };
        },
    };
    return {
        middleware,
        plugin,
    };
};
exports.default = costAnalysisTool;
//# sourceMappingURL=cost-analysis.tool.js.map